#include "header\GBA.h"
#include "header\BG.h"
#include "header\Timer.h"
#include "header\DMA.h"
#include "header\IRQ.h"
#include "header\sound.h"
#include "header\sprite.h"
#include "header\Blend.h"

#include "lib.h"
#include "iwram.h"
#include "irq.h"
#include "syscall.h"

#include "data.h"
#include "resource\BitMapFont.h"
#include "resource\BGFont.h"
#include <stdlib.h>

//const int __gba_multiboot;  //GBANp

#define BLOCK_X 56
#define BLOCK_Y 8
#define BLOCK_CX 32
#define BLOCK_CY 32
#define GRAPHIC_CX 128
#define GRAPHIC_CY 128
#define GRAPHIC_BLOCK_CX  (GRAPHIC_CX / BLOCK_CX)
#define GRAPHIC_BLOCK_CY  (GRAPHIC_CY / BLOCK_CY)
#define GRAPHIC_BLOCK_CNT (GRAPHIC_BLOCK_CX * GRAPHIC_BLOCK_CY)

#define BLOCK_MOVE_NUM 15

#define GAME_LEVEL     10
//---------------------------------------------------------------------------
typedef struct {
	u8 x;
	u8 y;
	u8 targetX;
	u8 targetY;
	u8 blockNo;
	u16 tileNo;
}ST_BLOCK;

ST_BLOCK BLOCK[GRAPHIC_BLOCK_CNT];		//15+1ubN

u8 MoveBlockPosition;	//ړłubN̈ʒu
u8 BGChr;
u8 BGMap;
//---------------------------------------------------------------------------
void Init();
void InitTimer();

void Title();
void Input();
void Random();
void Clear();

void MoveBlock();
void SwapBlock(u8);

u8 GetRndVect(s8);

u8 ChkVect(s8, s8, s8);
u8 ChkClear();

void IRQ_USER_Handler();
void IRQ_USER_Setup();
//---------------------------------------------------------------------------
int main(void)
{
	Init();
	InitTimer();

	REG_DISPCNT  = MODE_0 | BG1_ENABLE;
	Title();

	REG_DISPCNT |= OBJ_ENABLE | OBJ_MAP_2D;
	Random();

	for(;;){
		Input();

		if(ChkClear())
			break;
	}

	Clear();

	for(;;){}

	return 0;
}
//---------------------------------------------------------------------------
void Init(){

	MoveBlockPosition = BLOCK_MOVE_NUM;
	BGChr = 1;
	BGMap = 0;

	int i,tmpTileNo;
	for(i=0; i<GRAPHIC_BLOCK_CNT; i++){
		tmpTileNo  = (i % GRAPHIC_BLOCK_CX) * 8;		//x
		tmpTileNo += (i / GRAPHIC_BLOCK_CY) * 8 * GRAPHIC_BLOCK_CX * GRAPHIC_BLOCK_CY;	//y

		BLOCK[i].tileNo  = tmpTileNo;
		BLOCK[i].blockNo = i;
		BLOCK[i].x       = (SCREEN_CX/2) - BLOCK_CX/2;
		BLOCK[i].y       = (SCREEN_CY/2) - BLOCK_CY/2;
		BLOCK[i].targetX = BLOCK_X + (i % GRAPHIC_BLOCK_CX)*BLOCK_CX;
		BLOCK[i].targetY = BLOCK_Y + (i / GRAPHIC_BLOCK_CY)*BLOCK_CY;
	}

	//Sprite
	SpriteInitialize();
	SpriteSetPal256((u16*)EtfsGetDataOffset(_PHOTO_IMG_PAL_OFFSET));
	SpriteSetData((u16*)EtfsGetDataOffset(_PHOTO_IMG_BIN_OFFSET), _PHOTO_IMG_BIN_SIZE);

	for(i=0; i<GRAPHIC_BLOCK_CNT; i++){
		SpriteSetSize(i, SP_SIZE_32, SP_SQUARE, SP_COLOR_256);
		SpriteSetChr(i, BLOCK[i].tileNo);
	}

	//BG
	BGSetPal16((u16*)BGFont_palette);
	BGSetChrData(BGChr, (u16*)BGFont_data, BGFont_DATA_SIZE);

	REG_BG1CNT = (BG_SIZEA_512_512   | BG_COLOR_16       |
				  BG_CHARBASE(BGChr) | BG_MAPBASE(BGMap) | 0); 

	//wi 
	u16 x,y;
	for(y=0;y<32;y++){
		for(x=0;x<32;x++){
			BGDrawChar(BGMap,x,y,0,0x20);
		}
	}
}
//---------------------------------------------------------------------------
void InitTimer(){
	IRQ_USER_Setup();

	REG_TM0D = 0x0000;
	REG_TM0CNT = TM_ENABLE | TM_FREQ_PER_1 | TM_USEIRQ;
}
//========================================================================
void IRQ_USER_Setup(){
	REG_IME = IRQ_MASTER_OFF;

	REG_INTERUPT = (u32)IRQ_USER_Handler;
	REG_IE |= IRQ_BIT_TIMER0;

	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void IRQ_USER_Handler(){
	REG_IME  = IRQ_MASTER_OFF;
	u16 Flag = REG_IF;

	if( Flag & IRQ_BIT_TIMER0 ){
		MoveBlock();
	}

	REG_IF  = Flag;
	REG_IME = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
void MoveBlock(){
	u8 i;

	for(i=0; i<GRAPHIC_BLOCK_CNT; i++){

		if     (BLOCK[i].x > BLOCK[i].targetX) BLOCK[i].x--;
		else if(BLOCK[i].x < BLOCK[i].targetX) BLOCK[i].x++;

		if     (BLOCK[i].y > BLOCK[i].targetY) BLOCK[i].y--;
		else if(BLOCK[i].y < BLOCK[i].targetY) BLOCK[i].y++;

		SpriteMove(i, BLOCK[i].x, BLOCK[i].y);
	}
}
//---------------------------------------------------------------------------
void Title(){
	REG_TM1CNT = 0x00C0; //_p

	BGDrawStr(BGMap,1, 1,"Push A Button");
	BGDrawStr(BGMap,1,17,"15 Puzzle Sample. --");
	BGDrawStr(BGMap,1,18,"2005/05/15 by Akkera.    ---");
	WaitPushA();

	BGDrawStr(BGMap,1,1, "            ");

	REG_TM1CNT = 0x0000; //^C}[~߂
	srand(REG_TM1D);
}
//---------------------------------------------------------------------------
void Random(){
	u8 i;
	u8 cnt = rand()%GAME_LEVEL + 10;
	u8 pos = MoveBlockPosition;

	for(i=0; i<cnt; i++){
		pos = GetRndVect(pos);
		SwapBlock(pos);
	}

	MoveBlockPosition = pos;
}
//---------------------------------------------------------------------------
void SwapBlock(u8 num){
	u8 i;
	for(i=0; i<GRAPHIC_BLOCK_CNT; i++){
		if(BLOCK[i].blockNo == num)
			break;
	}

	u8 tmp1, tmp2, tmp3;

	tmp1 = BLOCK[i].targetX;
	tmp2 = BLOCK[i].targetY;
	tmp3 = BLOCK[i].blockNo;

	BLOCK[i].targetX = BLOCK[BLOCK_MOVE_NUM].targetX;
	BLOCK[i].targetY = BLOCK[BLOCK_MOVE_NUM].targetY;
	BLOCK[i].blockNo = BLOCK[BLOCK_MOVE_NUM].blockNo;

	BLOCK[BLOCK_MOVE_NUM].targetX = tmp1;
	BLOCK[BLOCK_MOVE_NUM].targetY = tmp2;
	BLOCK[BLOCK_MOVE_NUM].blockNo = tmp3;
}
//---------------------------------------------------------------------------
u8 GetRndVect(s8 posMas){
	s8 vect, pos;

ReLoad:

	pos  = posMas;
	vect = rand() % 4;

	switch(vect){
		case 0: pos -= GRAPHIC_BLOCK_CX; break;
		case 1: pos += GRAPHIC_BLOCK_CX; break;
		case 2: pos--;                   break;
		case 3: pos++;                   break;
	}

	//ړłH
	if( ChkVect(pos, posMas, vect) )
		goto ReLoad;

	return pos;
}
//---------------------------------------------------------------------------
u8 ChkVect(s8 pos, s8 posMas, s8 vect){

	if(pos<0 || pos>GRAPHIC_BLOCK_CNT-1)
		return 1;

	//PzȂ񂾂EEEi
	if((posMas+0) % GRAPHIC_BLOCK_CX == 0 && vect == 2)
		return 1;
	if((posMas+1) % GRAPHIC_BLOCK_CX == 0 && vect == 3)
		return 1;

	return 0;
}
//---------------------------------------------------------------------------
void Input(){
	s8 pos,vect;
	u32 key;

ReInput:

	for(;;){
		pos = MoveBlockPosition;
		key = *KEYS ^ 0x3ff;

		if(key & KEY_UP  ){
			pos -= GRAPHIC_BLOCK_CX;
			vect = 0;
			break;
		}
		if(key & KEY_DOWN ){
			pos += GRAPHIC_BLOCK_CX;
			vect = 1;
			break;
		}
		if(key & KEY_LEFT ){
			pos--;
			vect = 2;
			break;
		}
		if(key & KEY_RIGHT){
			pos++;
			vect = 3;
			break;
		}
	}//for(;;)

	if( ChkVect(pos, MoveBlockPosition, vect) )
		goto ReInput;

	SwapBlock(pos);
	MoveBlockPosition = pos;

	while(*KEYS != 0x3ff){}		//͂ɂȂ܂ő҂
}
//---------------------------------------------------------------------------
u8 ChkClear(){
	int i;
	for(i=0; i<GRAPHIC_BLOCK_CNT; i++){
		if(BLOCK[i].blockNo != i)
			return 0;
	}

	return 1;
}
//---------------------------------------------------------------------------
void Clear(){
	BGDrawStr(BGMap,1,1,"Clear!");
}
